@@ -3,6 +3,7 @@
 
     This file is part of Kleopatra, the KDE keymanager
     Copyright (c) 2007,2008 Klarälvdalens Datakonsult AB
+                  2018 Intevation GmbH
 
     Kleopatra is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -40,6 +41,7 @@
 #include "kleo/dn.h"
 #include "utils/filesystemwatcher.h"
 
+#include <gpgme++/gpgmepp_version.h>
 #include <gpgme++/error.h>
 #include <gpgme++/key.h>
 #include <gpgme++/decryptionresult.h>
@@ -48,6 +50,7 @@
 
 #include <qgpgme/protocol.h>
 #include <qgpgme/listallkeysjob.h>
+#include <qgpgme/cryptoconfig.h>
 
 #include <gpg-error.h>
 
@@ -87,7 +90,7 @@ class KeyCache::Private
     friend class ::Kleo::KeyCache;
     KeyCache *const q;
 public:
-    explicit Private(KeyCache *qq) : q(qq), m_refreshInterval(1), m_initalized(false)
+    explicit Private(KeyCache *qq) : q(qq), m_refreshInterval(1), m_initalized(false), m_pgpOnly(true)
     {
         connect(&m_autoKeyListingTimer, &QTimer::timeout, q, [this]() { q->startKeyListing(); });
         updateAutoKeyListingTimer();
@@ -195,6 +198,53 @@ public:
 
     void ensureCachePopulated() const;
 
+    void updateGroupCache() {
+
+#if GPGMEPP_VERSION < 0x10900 // 1.9.0
+        // In GPGME before 1.9.0 string lists are
+        // broken from cryptoconfig.
+        return;
+#else
+        // Update Group Keys
+        // this is a quick thing as it only involves reading the config
+        // so no need for a job.
+
+        // According to Werner Koch groups are more of a hack to solve
+        // a valid usecase (e.g. several keys defined for an internal mailing list)
+        // that won't make it in the proper keylist interface. And using gpgconf
+        // was the suggested way to support groups.
+        auto conf = QGpgME::cryptoConfig();
+        if (!conf) {
+            return;
+        }
+
+        auto entry = conf->entry(QStringLiteral("gpg"),
+                                 QStringLiteral("Configuration"),
+                                 QStringLiteral("group"));
+        if (!entry) {
+            return;
+        }
+
+        m_groups.clear();
+        for (const auto &value: entry->stringValueList()) {
+            const auto split = value.split(QLatin1Char('='));
+            if (split.size() != 2) {
+                qCDebug (LIBKLEO_LOG) << "Ignoring invalid group config:" << value;
+                continue;
+            }
+            const auto name = split[0];
+            const auto key = q->findByFingerprint(split[1].toLatin1().constData());
+            if (key.isNull()) {
+                qCDebug (LIBKLEO_LOG) << "Ignoring unknown fingerprint:" << split[1] << "in group" << name;
+                continue;
+            }
+            auto vec = m_groups.value(name);
+            vec.push_back(key);
+            m_groups.insert(name, vec);
+        }
+#endif
+    }
+
 private:
     QPointer<RefreshKeysJob> m_refreshJob;
     std::vector<std::shared_ptr<FileSystemWatcher> > m_fsWatchers;
@@ -207,6 +257,8 @@ private:
         std::vector<Subkey> subkeyid;
     } by;
     bool m_initalized;
+    bool m_pgpOnly;
+    QMap <QString, std::vector<Key> > m_groups;
 };
 
 std::shared_ptr<const KeyCache> KeyCache::instance()
@@ -294,6 +346,7 @@ void KeyCache::Private::refreshJobDone(const KeyListResult &result)
 {
     q->enableFileSystemWatcher(true);
     m_initalized = true;
+    updateGroupCache();
     Q_EMIT q->keyListingDone(result);
 }
 
@@ -668,6 +721,11 @@ std::vector<Key> KeyCache::findIssuers(std::vector<Key>::const_iterator first, s
 
 static std::string email(const UserID &uid)
 {
+    // Prefer the gnupg normalized one
+    const std::string addr = uid.addrSpec();
+    if (!addr.empty()) {
+        return addr;
+    }
     const std::string email = uid.email();
     if (email.empty()) {
         return DN(uid.id())[QStringLiteral("EMAIL")].trimmed().toUtf8().constData();
@@ -928,6 +986,7 @@ void KeyCache::insert(const std::vector<Key> &keys)
     by_chainid.swap(d->by.chainid);
 
     for (const Key &key : qAsConst(sorted)) {
+        d->m_pgpOnly &= key.protocol() == GpgME::OpenPGP;
         Q_EMIT added(key);
     }
 
@@ -1132,6 +1191,110 @@ void KeyCache::Private::ensureCachePopulated() const
     }
 }
 
+bool KeyCache::pgpOnly() const
+{
+    return d->m_pgpOnly;
+}
+
+static bool keyIsOk(const Key k)
+{
+    return !k.isExpired() && !k.isRevoked() && !k.isInvalid() && !k.isDisabled();
+}
+
+static bool uidIsOk(const UserID uid)
+{
+    return keyIsOk(uid.parent()) && !uid.isRevoked() && !uid.isInvalid();
+}
+
+static bool subkeyIsOk(const Subkey s)
+{
+    return !s.isRevoked() && !s.isInvalid() && !s.isDisabled();
+}
+
+std::vector<GpgME::Key> KeyCache::findBestByMailBox(const char *addr, GpgME::Protocol proto,
+                                                    bool needSign, bool needEncrypt) const
+{
+    d->ensureCachePopulated();
+    std::vector<GpgME::Key> ret;
+    if (!addr) {
+        return ret;
+    }
+    if (proto == Protocol::OpenPGP) {
+        // Groups take precendece over autmatic keys. We return all here
+        // so if a group key for example is expired we can show that.
+        ret = getGroupKeys(addr);
+        if (!ret.empty()) {
+            return ret;
+        }
+    }
+    Key keyC;
+    UserID uidC;
+    for (const Key &k: findByEMailAddress(addr)) {
+        if (proto != Protocol::UnknownProtocol && k.protocol() != proto) {
+            continue;
+        }
+        if (needEncrypt && !k.canEncrypt()) {
+            continue;
+        }
+        if (needSign && !k.canSign()) {
+            continue;
+        }
+        /* First get the uid that matches the mailbox */
+        for (const UserID &u: k.userIDs()) {
+            if (QString::fromStdString(u.addrSpec()).toLower() != QString::fromUtf8(addr).toLower()) {
+                continue;
+            }
+            if (uidC.isNull()) {
+                keyC = k;
+                uidC = u;
+            } else if ((!uidIsOk(uidC) && uidIsOk(u)) || uidC.validity() < u.validity()) {
+                /* Validity of the new key is better. */
+                uidC = u;
+                keyC = k;
+            } else if (uidC.validity() == u.validity() && uidIsOk(u)) {
+                /* Both are the same check which one is newer. */
+                time_t oldTime = 0;
+                for (const Subkey s: keyC.subkeys()) {
+                    if (!subkeyIsOk(s)) {
+                        continue;
+                    }
+                    if (needSign && s.canSign()) {
+                        oldTime = s.creationTime();
+                    } else if (needEncrypt && s.canEncrypt()) {
+                        oldTime = s.creationTime();
+                    } else if (s.canSign() || s.canEncrypt()) {
+                        oldTime = s.creationTime();
+                    }
+                }
+                time_t newTime = 0;
+                for (const Subkey s: k.subkeys()) {
+                    if (!subkeyIsOk(s)) {
+                        continue;
+                    }
+                    if (needSign && s.canSign()) {
+                        newTime = s.creationTime();
+                    } else if (needEncrypt && s.canEncrypt()) {
+                        newTime = s.creationTime();
+                    } else if (s.canSign() || s.canEncrypt()) {
+                        newTime = s.creationTime();
+                    }
+                }
+                if (newTime > oldTime) {
+                    uidC = u;
+                    keyC = k;
+                }
+            }
+        }
+    }
+    ret.push_back(keyC);
+    return ret;
+}
+
+std::vector<GpgME::Key> KeyCache::getGroupKeys(const QString &groupName) const
+{
+    return d->m_groups.value(groupName);
+}
+
 #include "moc_keycache_p.cpp"
 #include "moc_keycache.cpp"
 
